package nu.zoom.dsl.ast;

import nu.zoom.dsl.parser.StatesBaseVisitor;
import nu.zoom.dsl.parser.StatesParser;
import org.antlr.v4.runtime.tree.TerminalNode;

import java.util.*;
import java.util.stream.Stream;

public class StatesVisitorTransformer extends StatesBaseVisitor<StatesParser.DocumentContext> {
    private final HashMap<String,String> config = new HashMap<>();
    private final HashSet<TypeNode> nodeTypes = new HashSet<>();
    private final HashSet<TypeNode> messageTypes = new HashSet<>();
    // from -> <to, message>
    private final HashMap<String,HashMap<String, String>> transitions = new HashMap<>();

    @Override
    public StatesParser.DocumentContext visitTransition(StatesParser.TransitionContext ctx) {
        String from = ctx.from().state().typeName().IDENTIFIER().getText() ;
        String to = ctx.to().state().typeName().IDENTIFIER().getText() ;
        String message = ctx.message().typeName().IDENTIFIER().getText() ;
        this.transitions.computeIfAbsent(from, k -> new HashMap<>()).put(to, message);
        return super.visitTransition(ctx);
    }

    @Override
    public StatesParser.DocumentContext visitState(StatesParser.StateContext ctx) {
        String stateName = ctx.typeName().IDENTIFIER().getText() ;
        List<FieldNode> fields = extractFields(ctx.typeDeclaration()) ;
        this.nodeTypes.add(new TypeNode(stateName, fields));
        return super.visitState(ctx);
    }

    @Override
    public StatesParser.DocumentContext visitMessage(StatesParser.MessageContext ctx) {
        String messageName = ctx.typeName().IDENTIFIER().getText() ;
        List<FieldNode> fields = extractFields(ctx.typeDeclaration()) ;
        this.messageTypes.add(new TypeNode(messageName, fields));
        return super.visitMessage(ctx);
    }

    @Override
    public StatesParser.DocumentContext visitConfigitem(StatesParser.ConfigitemContext ctx) {
        String configKey = ctx.configkey().IDENTIFIER().getText();
        String configValue = getText(ctx.configvalue().IDENTIFIER(), ctx.configvalue().VALUE());
        this.config.put(configKey, configValue);
        return super.visitConfigitem(ctx);
    }

    public Set<StateNode> getStates() {
        HashSet<StateNode> states = new HashSet<>();
        this.transitions.forEach((state,v)->{
            HashSet<TransitionNode> transitionNodes = new HashSet<>();
            v.forEach((to, message) -> transitionNodes.add(new TransitionNode(message, to)));
            states.add(new StateNode(state, "", transitionNodes)) ;
        }) ;
        return states ;
    }

    public Map<String,String> getConfig() {
        return Map.copyOf(config);
    }

    public Set<TypeNode> getTypes() {
        final HashMap<String, TypeNode> stateTypeNodes = new HashMap<>();
        this.nodeTypes.forEach(typeNode -> {
            if (stateTypeNodes.containsKey(typeNode.name())) {
                TypeNode mergedNode = mergeTypeFields(typeNode, stateTypeNodes.get(typeNode.name()));
                stateTypeNodes.put(typeNode.name(), mergedNode);
            } else {
                stateTypeNodes.put(typeNode.name(), typeNode);
            }
        }) ;
        final HashMap<String, TypeNode> messageTypeNodes = new HashMap<>();
        this.messageTypes.forEach(typeNode -> {
            if (stateTypeNodes.containsKey(typeNode.name())) {
                throw new ParseException("Message " + typeNode.name() + " conflicts with state with the same type name");
            }
            if (messageTypeNodes.containsKey(typeNode.name())) {
                TypeNode mergedNode = mergeTypeFields(typeNode, messageTypeNodes.get(typeNode.name()));
                messageTypeNodes.put(typeNode.name(), mergedNode);
            } else {
                messageTypeNodes.put(typeNode.name(), typeNode);
            }
        }) ;
        HashSet<TypeNode> allTypeNodes = new HashSet<>(stateTypeNodes.values());
        allTypeNodes.addAll(messageTypeNodes.values());
        return allTypeNodes ;
    }

    private TypeNode mergeTypeFields(TypeNode t1, TypeNode t2) {
        List<FieldNode> t1Fields = (t1 != null) ? t1.fields() : List.of() ;
        List<FieldNode> t2Fields = (t2 != null) ? t2.fields() : List.of() ;
        HashMap<String, FieldNode> mergedFields = new HashMap<>();
        t1Fields.forEach(field -> {
            if (mergedFields.containsKey(field.name())) {
                throw new ParseException("Duplicate field name: " + field.name());
            }
            mergedFields.put(field.name(), field);
        });
        t2Fields.forEach(field -> {
            if (mergedFields.containsKey(field.name())) {
                throw new ParseException("Duplicate field name: " + field.name());
            }
            mergedFields.put(field.name(), field);
        });
        return new TypeNode(t1.name(), mergedFields.values().stream().toList()) ;
    }

    private List<FieldNode> extractFields(StatesParser.TypeDeclarationContext declaration) {
        if (declaration == null) {
            return Collections.emptyList();
        }
        return declaration
                .typeField()
                .stream()
                .map(
                        ctx ->
                                new FieldNode(ctx.fieldName().getText(), ctx.fieldType().getText())
                )
                .toList();
    }

    // Concatenate the text from two terminal nodes. Useful for contexts that are either an identifier or a value,
    // and you just want the text from whichever is not null.
    private String getText(TerminalNode identifier, TerminalNode value) {
        return
                ((identifier != null) ? identifier.getText() : "") +
                        ((value != null) ? value.getText() : "");
    }

}
