package nu.zoom.dsl.run;

import freemarker.template.TemplateException;
import nu.zoom.dsl.ast.DocumentNode;
import nu.zoom.dsl.freemarker.Generator;

import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.List;
import java.util.Optional;

public final class Runner {
    public static final String DEFAULT_TEMPLATE_DIRECTORY_NAME = "endpoints-template";
    public static final String DEFAULT_OUTPUT_DIRECTORY_NAME = "endpoints-output";

    public static void run(
            Path dsl,
            Optional<Path> templates,
            Optional<Path> output,
            Optional<ParserType> maybeParser,
            Logger logger
    ) throws ValidationException, IOException, GeneratorException {
        Path templatesDir = templates.orElse(Paths.get(DEFAULT_TEMPLATE_DIRECTORY_NAME));
        Path outputDir = output.orElse(Paths.get(DEFAULT_OUTPUT_DIRECTORY_NAME));

        validateOutputDirectory(outputDir);
        validateTemplateDirectory(templatesDir);
        validateInputFile(dsl);
        logger.println("Parsing: " + dsl.toAbsolutePath());

        final ParserType parser =
                maybeParser.orElseGet(
                        () -> {
                            if (dsl.getFileName().toString().endsWith(".states")) {
                                return ParserType.States;
                            } else {
                                return ParserType.Endpoints;
                            }
                        }
                );

        final DocumentNode rootNode;
        if (parser == ParserType.States) {
            logger.println("using state grammar.");
            rootNode = ParserWrapper.parseStates(dsl);
        } else {
            logger.println("using endpoints grammar.");
            rootNode = ParserWrapper.parseEndpoints(dsl);
        }
        logger.println("AST: " + rootNode);
        logger.println("Generating from templates in: " + templatesDir.toAbsolutePath());
        Generator generator = new Generator(templatesDir, rootNode, outputDir);
        List<Path> generatedPaths = null;
        try {
            generatedPaths = generator.generate();
        } catch (TemplateException e) {
            throw new GeneratorException(e);
        }
        if (generatedPaths.isEmpty()) {
            System.out.println("No generated paths found.");
        } else {
            generatedPaths.forEach(p -> logger.println("Generated: " + p.toAbsolutePath()));
        }
    }

    private static void validateOutputDirectory(Path outputDir) throws IOException, ValidationException {
        if (Files.notExists(outputDir)) {
            Files.createDirectories(outputDir);
        }
        if (!Files.isDirectory(outputDir)) {
            throw new ValidationException("Output directory: '" + outputDir + " 'is not a directory.");
        }
    }

    private static void validateTemplateDirectory(Path templateDir) throws ValidationException {
        if (!Files.isDirectory(templateDir)) {
            throw new ValidationException("Template directory '" + templateDir + "' is not a directory.");
        }
    }

    private static void validateInputFile(Path file) throws ValidationException {
        if (Files.notExists(file)) {
            throw new ValidationException("Input file '" + file + "' does not exist.");
        }
        if (!Files.isReadable(file) || !Files.isRegularFile(file)) {
            throw new ValidationException("Input file '" + file + "' is not a readable file.");
        }
    }

    public enum ParserType {
        Endpoints,
        States
    }
}
